
using System.Collections.Generic;
using Jypeli;
namespace matopeli;

/// @author Paulus Kopperoinen
/// @version 30.10.2023
/// <summary>
/// Perinteinen matopeli, jossa tavoitteena kasvattaa matoa mahdollisimman suureksi omenia syömällä.
/// </summary>

public class Matopeli : PhysicsGame
{
    private Objektit mato;
    private Objektit omena;
    private Objektit matomuu;

    private IntMeter Pistelaskuri;
    public IntMeter Ennatyslaskuri;

    private readonly List<Vector> MatoSijainti = new List<Vector>();
    private readonly List<PhysicsObject> MatoKeho = new List<PhysicsObject>();

    private int Leveys;
    private int KenttaKoko;
    private double Nopeus; 
    private int KentanSijainti;
    
    private static readonly Image Seinakuva = LoadImage("seina.png");
    private static readonly Image Matopaakuva = LoadImage("matopaa.png");
    private readonly Image matokehokuva = LoadImage("matokeho.png");
    private readonly Image omenakuva = LoadImage("omena.png");
    
    private bool suuntaYlos;
    private bool suuntaAlas;
    private bool suuntaOikea;
    private bool suuntaVasen;

    private bool matoLiikkunut;
    private bool omenaKentalla;
    
    
    /// <summary>
    /// Kun peli käynnistetään ensimmäisen kerran
    /// </summary>
    public override void Begin()
    {
        AloitaUusiPeli();
    }

    
    /// <summary>
    /// Poistaa vanhan pelin tiedot ja alustaa uuden pelin
    /// </summary>
    private void AloitaUusiPeli()
    {
        MatoKeho.Clear();
        MatoSijainti.Clear();
        ClearAll();
        Asetukset();
        LuoMato(Luoseinat(KenttaKoko, Leveys));
        Asetaohjaimet();
        LuoAikalaskuri();
        LuoPistelaskuri();
        LuoOmena();
        Level.Background.Color = Color.LightGreen;
    }
    
    
    /// <summary>
    /// Määrittää ohjaimet, joita pelaaja voi käyttää ja mitä niitä painamalla tapahtuu
    /// </summary>
    private void Asetaohjaimet()
    {
        Keyboard.Listen(Key.Up, ButtonState.Pressed, VaihdaSuuntaYlos, "Turn upwards");
        Keyboard.Listen(Key.Down, ButtonState.Pressed, VaihdaSuuntaAlas, "Turn downwards");
        Keyboard.Listen(Key.Left, ButtonState.Pressed, VaihdaSuuntaVasen , "Turn left");
        Keyboard.Listen(Key.Right, ButtonState.Pressed, VaihdaSuuntaOikea, "Turn right");
        Keyboard.Listen(Key.Enter, ButtonState.Pressed, LiikutaMatoa, "New Game");
    }


    /// <summary>
    /// Laskee pisteet, joita kerätään omenoita syömällä
    /// </summary>
    private void LuoPistelaskuri()
    {
        Pistelaskuri = new IntMeter(0);               
        Label pistenaytto = new Label(); 
        pistenaytto.X = KentanSijainti+ KenttaKoko/2  * Leveys;
        pistenaytto.Y = KentanSijainti + (KenttaKoko+2) * Leveys;
        pistenaytto.TextColor = Color.Brown;
        pistenaytto.Color = Color.LightGreen;
        pistenaytto.Title = "Pisteet: ";
        pistenaytto.BindTo(Pistelaskuri);
        Add(pistenaytto);
    }
    

    /// <summary>
    /// Pelin asetukset
    /// </summary>
    private void Asetukset()
    {
    Leveys = 40;
    KenttaKoko = 16;
    KentanSijainti = -300;
    suuntaYlos = false;
    suuntaAlas = false;
    suuntaOikea = true; 
    suuntaVasen = false;
    Nopeus = 0.1;
    matoLiikkunut = false;
    omenaKentalla = true;
    }
    
    
    /// <summary>
    /// Päättää pelin ja aloitaa uuden, kun mato kuolee
    /// </summary>
    private void MatoKuolee()
    {
        AloitaUusiPeli();
    }
    
    
    /// <summary>
    /// Luo madon kentälle
    /// </summary>
    /// <returns></returns>
    private void LuoMato(Vector paikka)
    {
        mato = new Objektit(Leveys-5, Leveys-5);
        mato.Image = Matopaakuva;
        Add(mato);
        mato.Tag = "mato";
        mato.Position = ArvoPaikka(paikka,3);
        AddCollisionHandler(mato, "omena", delegate(IPhysicsObject omena, IPhysicsObject mato) {MatoSyoOmenan();  });
        AddCollisionHandler(mato, "seina", delegate(IPhysicsObject seina, IPhysicsObject mato) {MatoKuolee();  });
        AddCollisionHandler(mato, "matokeho", delegate(IPhysicsObject matomuu, IPhysicsObject mato) {MatoKuolee();  });
    }

    
    /// <summary>
    /// Luodaan omena johonkin satunnaiseen kohtaan kartalla, joka ei ole samassa kohtaa madon kanssa.
    /// </summary>
    private void LuoOmena()
    {
        int i = 1;
        Vector paikka = ArvoPaikka(new Vector(KentanSijainti,KentanSijainti),0);
        while (i>0)
        {
            if (paikka != mato.Position)
            {
                break;
            }
            paikka = ArvoPaikka(new Vector(KentanSijainti,KentanSijainti),0);
        }
        omena = new Objektit(Leveys-10, Leveys-10);
        omena.Shape = Shape.Circle;
        omena.Position = paikka;
        omena.Image = omenakuva;
        omena.Tag = "omena";
        Add(omena);
        omenaKentalla = true;
    }

    
    /// <summary>
    /// Arpoo sattumanvaraisen paikan kentältä ja palauttaa kyseisen vektorin
    /// </summary>
    /// <param name="alkupaikka">Mihin arvottavaa paikkaa halutaan suhteuttaa</param>
    /// <param name="Keskitys">Kuinka paljon halutaan saada arvottavan paikan maksimaalinen arvo pois reunoilta</param>
    /// <returns></returns>
    private Vector ArvoPaikka(Vector alkupaikka, int Keskitys)
    {
        Vector Paikka;
        double x;
        double y;
        Vector omenaxy0 = new Vector(alkupaikka.X + Leveys, alkupaikka.X);
        
        y = RandomGen.NextInt(1 + Keskitys, KenttaKoko-1-Keskitys);
        y = omenaxy0.Y + y * Leveys;
            
        x = RandomGen.NextInt(1+Keskitys, KenttaKoko-1-Keskitys);
        x = omenaxy0.X + x * Leveys;
        
        Paikka = new Vector(x, y);
        return Paikka;
    }
        
    
    /// <summary>
    /// Luo kentän reunat
    /// </summary>
    /// <param name="maara">luotavan kentän yhtä reunaa varten luotavien palikoiden määrä</param>
    /// <param name="koko">kuinka isoja palikoita tehdään</param>
    private Vector Luoseinat(int maara, int koko)
    {
        double x = KentanSijainti;
        double y = KentanSijainti;
        for (int i = 0; i < maara; i++)
        {
            LuoSeina(koko,new Vector(x,y));
            x += koko;
        }
        for (int i = 0; i < maara+1; i++)
        {
            LuoSeina(koko,new Vector(x,y));
            y += koko;
        }
        for (int i = 0; i < maara+1; i++)
        {
            LuoSeina(koko,new Vector(x,y));
            x -= koko;
        }
        for (int i = 0; i < maara+2; i++)
        {
            LuoSeina(koko,new Vector(x,y));
            y -= koko;
        }

        return new Vector(x, y);
    }
    
    
    /// <summary>
    /// Luo kentän palikat, joista kentän reunat muodostuvat
    /// </summary>
    private void LuoSeina(int leveys, Vector paikka)
    {
        PhysicsObject seina = new PhysicsObject(leveys,leveys);
        seina.MakeStatic();
        seina.Color = Color.Green;
        seina.Tag = "seina";
        seina.Position = paikka;
        seina.Image = Seinakuva;
        Add(seina);
    }
    
    
    /// <summary>
    /// Tehdään aliohjelma, joka luo uusia osia madon perään.
    /// </summary>
    private void LuoMatoKeho()
    {
        matomuu = new Objektit(Leveys-6, Leveys-6);
        matomuu.Tag = "matokeho";
        matomuu.Image = matokehokuva;
        Vector paikka = MatoSijainti[MatoSijainti.Count -1 - MatoKeho.Count];
        matomuu.Position = paikka;
        Add(matomuu);
        MatoKeho.Add(matomuu);
        AddCollisionHandler(matomuu, "omena", delegate(IPhysicsObject mato, IPhysicsObject omena) {LuoUusiOmena();  });
    }

    
    /// <summary>
    /// Jos omena luodaan madon kehon sisään, tämä aliohjelma luo uuden omenan vanhan tilalle
    /// </summary>
    private void LuoUusiOmena()
    {
        omena.Destroy();
        omenaKentalla = false;
    }
    
    
    /// <summary>
    /// Tekee tarvittavat toimet, kun mato onnistuu syömään omenan
    /// </summary>
    private void MatoSyoOmenan()
    {
        omena.Destroy();
        Pistelaskuri.Value += 1;
        omenaKentalla = false;
    }
    
    
    /// <summary>
    /// Antaa tiedon LiikutaMatoa aliohjelmalle, että madon pitäisi kääntyä ylös.
    /// </summary>
     private void VaihdaSuuntaYlos()
        {
            
            if (suuntaAlas)
            {
                suuntaOikea = false;
                suuntaVasen = false;
                suuntaYlos = false;
                suuntaAlas = true;
            }
            else
            {
                suuntaOikea = false;
                suuntaVasen = false;
                suuntaYlos = true;
                suuntaAlas = false;
            }
        }
     
    
    /// <summary>
    /// Antaa tiedon LiikutaMatoa aliohjelmalle, että madon pitäisi kääntyä alas.
    /// </summary>
     private void VaihdaSuuntaAlas()
        {
            if (suuntaYlos)
            {
                suuntaOikea = false;
                suuntaVasen = false;
                suuntaYlos = true;
                suuntaAlas = false;
            }
            else
            {
                suuntaOikea = false;
                suuntaVasen = false;
                suuntaYlos = false;
                suuntaAlas = true;
            }
        }
      
     
    /// <summary>
    /// Antaa tiedon LiikutaMatoa aliohjelmalle, että madon pitäisi kääntyä oikeaan.
    /// </summary>
     private void VaihdaSuuntaOikea()
        {
            if (suuntaVasen)
            {
                
                suuntaOikea = false;
                suuntaVasen = true;
                suuntaYlos = false;
                suuntaAlas = false;
            }
            else
            {
                suuntaOikea = true;
                suuntaVasen = false;
                suuntaYlos = false;
                suuntaAlas = false;
            }

        }
       
     
    /// <summary>
    /// Antaa tiedon LiikutaMatoa aliohjelmalle, että madon pitäisi kääntyä vasempaan.
    /// </summary>
     private void VaihdaSuuntaVasen()
        {
            if (suuntaOikea)
            {
                suuntaOikea = true;
                suuntaVasen = false;
                suuntaYlos = false;
                suuntaAlas = false;
            }
            else
            {
                suuntaOikea = false;
                suuntaVasen = true;
                suuntaYlos = false;
                suuntaAlas = false;
            }
        }
     
     
     /// <summary>
     /// Aikalaskuri toimii pelin moottorina ja antaa PaivitaKentälle tiedon, kun pitää tarkistaa mitä kentällä on tapahtunut.
     /// </summary>
     private void LuoAikalaskuri()
        {
            Timer aikalaskuri = new Timer();
            aikalaskuri.Interval=Nopeus;
            aikalaskuri.Start();
            aikalaskuri.Timeout += PaivitaKentta; 
        }

     
     /// <summary>
     /// Kun ajastimesta loppuu aika, siirtyy peli tähän funktioon, jonka tarkoituksena on hahmottaa, mitä kaikkea pelissä on tapahtunut ja miten siihen tulee reagoida.
     /// </summary>
     private void PaivitaKentta()
     {
         if (matoLiikkunut == false)
         {
             return;
         }
         
         if (omenaKentalla == false)
         {
             LuoMatoKeho();
             LuoOmena();
         }
         
         MatoSijainti.Add(new Vector(mato.X,mato.Y));
         LiikutaMatoa();
         LiikutaKehoa();
         MatoSijainti.Add(new Vector(mato.X,mato.Y));  
     }
     
     
     /// <summary>
     /// LiikutaMatoa aliohjelman tehtävä on selvittää tieto siitä, mihin suuntaan on tarkoitus liikkua ja liikuttaa matoa siihen suuntaan.
     /// </summary>
    private void LiikutaMatoa()
    {
        matoLiikkunut = false;
        if (suuntaYlos)
        {
            mato.Y = mato.Y + Leveys;
            mato.Angle = Angle.FromDegrees(90);
        }
        if (suuntaAlas)
        {
            mato.Y= mato.Y - Leveys;
            mato.Angle = Angle.FromDegrees(270);
        }
        if (suuntaOikea)
        {
            mato.X= mato.X + Leveys;
            mato.Angle = Angle.FromDegrees(0);
           
        }
        if (suuntaVasen)
        {
            mato.X= mato.X - Leveys;
            mato.Angle = Angle.FromDegrees(180);
            
        }
        matoLiikkunut = true;
    }

    
    /// <summary>
    /// Liikutakehoa, aliohjelman tehtävänä on liikuttaa madon kehoa sen pään mukaisesti.
    /// </summary>
    private void LiikutaKehoa()
    {
        if (MatoKeho.Count==0)
        {
            return;
        }
        matoLiikkunut = false;
        MatoKeho[0].X = MatoSijainti[MatoSijainti.Count - 1].X;
        MatoKeho[0].Y = MatoSijainti[MatoSijainti.Count - 1].Y;
        
        for (int i = 1; i < MatoKeho.Count; i++)
        {
            MatoKeho[i].X = MatoSijainti[MatoSijainti.Count - 2*i].X;
            MatoKeho[i].Y = MatoSijainti[MatoSijainti.Count - 2*i].Y;
        }

        matoLiikkunut = true;
    }
    
    
    /// <summary>
    /// Pelin fysiikkaobjektit luodaan tämän classin mukaisesti
    /// </summary>
    public class Objektit : PhysicsObject
    {
         public Objektit(double leveys, double korkeus)
        : base(leveys, korkeus)
        {
            Mass = 1000;
            IsUpdated = true;
            IgnoresCollisionResponse = true;
        }
    }
}